import localforage from 'localforage';
import axios from 'axios';
import queryString from 'query-string';

/**
 * Token auth class
 * @class TokenAuth
 */
export default class TokenAuth {
  constructor(options = {}) {
    this.options = Object.assign({}, options);
    this.user = {};

    this.globalAuthPromise = null;

    this.apiInstance = axios.create({
      baseURL: this.options.apiUrl,
    });
    this.localForageInstance = localforage
      .createInstance(this.options.localforage);
  }

  initializeDom() {
    if (typeof window !== 'undefined') {
      window.addEventListener('message', this.handlePostmessage, false);
    }

    if (this.options.validateOnPageLoad) {
      this.validateUser();
    }
  }

  /**
   * @method authenticate
   * @param {string} provider - A string that is also the name of the target
   * provider service.
   * @param {object} [options] - An object containing the following params
   */
  authenticate(provider, options = {}) {
    if (typeof provider === undefined || typeof provider !== 'string') {
      throw new Error('The authentication provider name must be specified.');
    }

    this.openAuthWindow(provider, options);
  }
  signout(redirectTo) {
    return this.apiInstance.delete(this.options.signOutUrl)
      .then((res) => {
        this.emitEvent('auth:logout-success', res);
      })
      .catch((err) => {
        this.emitEvent('auth:logout-error', err);
      })
      .then(() => {
        this.invalidateTokens().then(console.log);
        if (typeof window !== 'undefined') {
          window.location.href = '/';
        }
      });
  }
  /**
   * @method openAuthWindow
   * @param {string} provider - A string that is also the name of the target
   * provider service.
   * @param {object} [options] - An object containing the following params
   */
  openAuthWindow(provider, options = {}) {
    const { omniauthWindowType } = this.options;
    const authUrl = this.buildAuthUrl(omniauthWindowType, provider, options);
    if (omniauthWindowType === 'newWindow') {
      this.requestCredentialsViaPostMessage(this.createPopup(authUrl));
    } else if (omniauthWindowType === 'inAppBrowser') {
      // this.requestCredentialsViaExecuteScript(this.createPopup(authUrl));
    } else if (omniauthWindowType === 'sameWindow') {
      this.visitUrl(authUrl);
    } else {
      throw new Error('Unsupported omniauthWindowType');
    }
  }
  visitUrl(url) {
    return typeof window !== 'undefined' ? window.location.replace(url) : null;
  }
  createPopup(url) {
    return typeof window !== 'undefined' ?
      window.open(url, '_blank', 'closebuttoncaption=Cancel') :
      null;
  }
  requestCredentialsViaPostMessage(authWindow) {
    console.log('got some response here?', authWindow);
    if (authWindow.closed) {
      // this.handleAuthWindowClose(authWindow);
    } else {
      authWindow.postMessage('requestCredentials', '*');
      // this.requestCredentialsPollingTimer = setTimeout(() => {
      //   this.requestCredentialsViaPostMessage(authWindow);
      // }, 500);
    }
  }
  handleValidAuth(user, setHeader = false) {
    return new Promise((resolve) => {
      console.log('handle valid auth data', user);
      this.user = Object.assign({}, this.user, user);
  
      this.user.signedIn = true;
      if (setHeader) {
        this.setAuthHeaders({
          'Access-Token': this.user.auth_token,
          'Token-Type': 'Baerer',
          Client: this.user.client_id,
          Uid: this.user.uid,
          Expiry: this.user.expiry,
        });
      }

      resolve();
    });
  }
  userIsAuthenticated() {
    return new Promise(async (resolve, reject) => {
      const authHeaders = await this.retrieveData('auth_headers');
      if (authHeaders && authHeaders['Access-Token']) {
        console.log('got access token', authHeaders['Access-Token']);
        console.log('user?', this.user);
        if (this.user.signedIn) {
          const expired = await this.tokenHasExpired();
          console.log('expired?', expired);
        }
        resolve(true);
      } else {
        reject();
      }
    });
  }
  /**
   * @method persistData
   */
  persistData(key, value) {
    return this.localForageInstance.setItem(key, value);
  }
  /**
   * @method retrieveData
   * @param {string} key - Data key name
   * @returns {Promise}
   */
  retrieveData(key) {
    try {
      if (!key) {
        throw new Error('A key is required to retrieve the data from the localforage');
      }

      this.haveLocalforage();
      return this.localForageInstance.getItem(key);
    } catch (e) {
      return Promise.reject(e);
    }
  }
  deleteData(key) {
    return this.localForageInstance.removeItem(key);
  }
  tokenHasExpired() {
    return new Promise((resolve, reject) => {
      const now = new Date().getTime();
      const expiry = this.getExpiry()
        .then((res) => {
          if (res && res < now) {
            resolve();
          } else {
            reject();
          }
        })
        .catch(() => reject());
    });
  }
  getExpiry() {
    return this.retrieveData('auth_headers')
      .then((res) => {
        const expiry = this.parseExpiry(res);
      });
  }
  /**
   * @method buildAuthUrl
   * @param {string} windowType - The omniauthwindow type to use
   * @param {string} provider - A string that is also the name of the target
   * provider service.
   * @param {object} [options] - An object containing the following params
   * @returns {string} authUrl - The generated auth url
   */
  buildAuthUrl(windowType, provider, options = {}) {
    const opts = Object.assign({}, this.options, options);
    const { apiUrl, authProviderPaths, confirmationSuccessUrl } = opts;
    const authProviderPath = authProviderPaths[provider];
    const originUrl = confirmationSuccessUrl;
    
    let authUrl = apiUrl;
    authUrl += authProviderPath;
    authUrl += `?auth_origin_url=${encodeURIComponent(originUrl)}`;

    /**
     * Create a list of params to inject in the auth URL
     */
    const params = Object.assign({}, options.params || {}, {
      omniauth_window_type: windowType,
    });
    
    /**
     * Inject all the created params in the auth URL
     */
    authUrl += Object.keys(params)
      .map((p) => `&${encodeURIComponent(p)}=${encodeURIComponent(params[p])}`)
      .join('');

    return authUrl;
  }
  /**
   * @method validateUser
   */
  validateUser() {
    return new Promise((resolve, reject) => {
      this.userIsAuthenticated()
        .then((res) => {
          console.log('authenticated?', res);
          resolve();
        })
        .catch((err) => {
          console.log('err?', err);
          const query = queryString(window.location.search);
          const token = query.auth_token;
          if (token) {
            this.setAuthHeaders({
              'Access-Token': query.auth_token,
              'Token-Type': 'Baerer',
              Client: query.client_id,
              Uid: query.uid,
              Expiry: query.expiry,
            })
            .then(() => {
              // Remove params from the url
              let url = (window.location.origin + window.location.pathname) || '/';
              let params = Object.assign({}, queryString(window.location.search));
              ['auth_token', 'blank', 'token', 'client_id', 'uid', 'expiry', 'config', 'reset_password', 'account_confirmation_success', 'oauth_registration'].forEach((prop) => {
                delete params[prop];
              });
  
              if (Object.keys(params).length > 0) {
                url += '?' + this.buildQueryString(params);
              }
              window.location.href = url;
        
              // this.nuxtContext.redirect('/');
              this.validateToken()
                .then((r) => resolve(r))
                .catch((e) => reject(e));
            });
      
          } else {
            if (this.options.forceValidationToken) {
              console.log('force validation');
              this.validateToken()
                .then((r) => resolve(r))
                .catch((e) => reject(e));
            } else {
              console.log('retrieve from storage');
              this.retrieveData('auth_headers')
                .then((res) => {
                  if (res) {
                    this.tokenHasExpired()
                      .then(() => {
                        console.log('token expired');
                        this.emitEvent('auth:session-expired');
                        reject({
                          reason: 'Unauthorized',
                          errors: [
                            'Session expired',
                          ],
                        });
                      })
                      .catch(() => {
                        console.log('check for token expired');
                        this.validateToken()
                          .then((r) => resolve(r))
                          .catch((e) => reject(e));
                      });
                  } else {
                    reject({
                      reason: 'Unauthorized',
                      errors: [
                        'No credentials',
                      ],
                    });
                  }
                  console.log('res from storage', res);
                })
                .catch(() => {
                  reject({
                    reason: 'Unauthorized',
                    errors: [
                      'No credentials',
                    ],
                  });
                  this.emitEvent('auth:invalid');
                });
            }
          }
        });
    });
  }
  buildQueryString(params) {
    let str = [];
    str = Object.keys(params)
      .filter((v, k) => v !== undefined && k !== undefined)
      .map((v, k) => `${k}=${encodeURIComponent(v)}`);
    return str.join('&');
  }
  invalidateTokens() {
    return new Promise((resolve, reject) => {
      Object.keys(this.user).forEach((k) => {
        if (this.user[k]) {
          delete this.user[k];
        }
      });
      console.log('user', this.user);

      this.deleteData('auth_headers')
        .then(resolve)
        .catch(reject);
    });
  }
  validateToken() {
    return this.apiInstance.get(this.options.tokenValidationPath)
      .then((resp) => {
        this.handleValidAuth(resp.data);
        console.log('response from api', resp);
      })
      .catch((err) => {
        console.log('error from api', err);
      });
  }
  /**
   * @method submitRegistration
   */
  submitRegistration() {
  }
  /**
   * @method submitLogin
   */
  submitLogin() {
  }
  /**
   * @method signOut
   */
  signOut() {
  }
  /**
   * @method requestPasswordReset
   */
  requestPasswordReset() {
  }
  /**
   * @method updatePassword
   */
  updatePassword() {
  }
  /**
   * @method updateAccount
   */
  updateAccount() {
  }
  /**
   * @method destroyAccount
   */
  destroyAccount() {
  }
  // buildAuthHeaders(context) {
  //   const tagRegex = /(\(token\)|\(clientId\))/;
  //   const tagInsideRegex = /\((\w+)\)/;

  //   tokenFormat: {
  //     'Access-Token': '(token)',
  //     'Token-Type': 'Bearer',
  //     Client: '(clientId)',
  //     Expiry: '(expiry)',
  //     Uid: '(uid)',
  //   },

  //   function interpolate(key, v) {

  //   }

  //   const headers = [];
  //   Object.keys(this.options.tokenFormat).forEach((key) => {
  //     const stringToBeReplaced = this.options.tokenFormat[key];
  //     const res = interpolate()      
  //   });
  //   Object.keys(context).forEach((k) => {
      
  //   });
    
  // }
  setAuthHeaders(headers) {
    console.log('set headers', headers);
    return new Promise((resolve, reject) => {
      let newHeaders = {};
      this.retrieveData('auth_headers')
        .then((res, err) => {
          if (res) {
            newHeaders = res;
          } else {
            newHeaders = headers;
          }

          // Set the headers in the api instance
          console.log('headers to set', newHeaders);

          this.persistData('auth_headers', newHeaders)
            .then((r) => {
              console.log('res from persistant storage', r);
              if (r) {
                Object.keys(newHeaders).forEach((k) => {
                  ['get', 'delete', 'patch', 'put', 'post', 'common', 'head'].forEach((h) => {
                    this.apiInstance.defaults.headers[h][k] = newHeaders[k];
                  });
                });
                resolve();
              } else {
                reject(error);
              }
            });
        });
    });
    const newHeaders = Object.assign({}, )
  }
  handlePostmessage(e) {
    if (e.data.message === 'deliverCredentials') {
      delete e.data.message;
      this.handleValidAuth(e.data, true);
      // Check if a new user was registered
    } else if (e.data.message === 'authFailure') {
      const error = {
        reason: 'unauthorized',
        errors: [
          e.data.error,
        ],
      };
    }
  }
  parseExpiry(headers) {
    const expiry = headers['expiry'];
    return parseInt(expiry, 10) * 1000 || null;
  }
  haveLocalforage() {
    if (this.localForageInstance === 'undefined' || this.localForageInstance === null) {
      throw new Error('LocalForage instance was not found. Plugin compatible only with async storage. Read more about localforage at https://github.com/localForage/localForage');
    }
  }
  emitEvent(eventName, val) {
    if (typeof window !== 'undefined') {
      if (window['Event']) {
        const event = new Event(eventName);
        window.dispatchEvent(event);
      }
    }
  }
}
